"""Data Handler using kz-filtered data."""

__author__ = 'Lukas Leufen'
__date__ = '2020-08-26'

import inspect
import numpy as np
import pandas as pd
import xarray as xr
from typing import List, Union, Tuple, Optional
from functools import partial

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler
from mlair.helpers import remove_items, to_list, TimeTrackingWrapper
from mlair.helpers.filter import KolmogorovZurbenkoFilterMovingWindow as KZFilter
from mlair.helpers.filter import FIRFilter, ClimateFIRFilter

# define a more general date type for type hinting
str_or_list = Union[str, List[str]]


# cutoff_p = [(None, 14), (8, 6), (2, 0.8), (0.8, None)]
# cutoff = list(map(lambda x: (1. / x[0] if x[0] is not None else None, 1. / x[1] if x[1] is not None else None), cutoff_p))
# fs = 24.
# # order = int(60 * fs) + 1
# order = np.array([int(14 * fs) + 1, int(14 * fs) + 1, int(4 * fs) + 1, int(2 * fs) + 1])
# print("cutoff period", cutoff_p)
# print("cutoff", cutoff)
# print("fs", fs)
# print("order", order)
# print("delay", 0.5 * (order-1) / fs)
# window = ("kaiser", 5)
# # low pass
# y, h = fir_filter(station_data.values.flatten(), fs, order[0], cutoff_low = cutoff[0][0], cutoff_high = cutoff[0][1], window=window)
# filtered = xr.ones_like(station_data) * y.reshape(station_data.values.shape)


class DataHandlerFilterSingleStation(DataHandlerSingleStation):
    """General data handler for a single station to be used by a superior data handler."""

    _requirements = remove_items(DataHandlerSingleStation.requirements(), "station")
    _hash = DataHandlerSingleStation._hash + ["filter_dim"]

    DEFAULT_FILTER_DIM = "filter"

    def __init__(self, *args, filter_dim=DEFAULT_FILTER_DIM, **kwargs):
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.filter_dim = filter_dim
        super().__init__(*args, **kwargs)

    def setup_transformation(self, transformation: Union[None, dict, Tuple]) -> Tuple[Optional[dict], Optional[dict]]:
        """
        Adjust setup of transformation because filtered data will have negative values which is not compatible with
        the log transformation. Therefore, replace all log transformation methods by a default standardization. This is
        only applied on input side.
        """
        transformation = super(__class__, self).setup_transformation(transformation)
        if transformation[0] is not None:
            for k, v in transformation[0].items():
                if v["method"] == "log":
                    transformation[0][k]["method"] = "standardise"
        return transformation

    def _check_sampling(self, **kwargs):
        assert kwargs.get("sampling") == "hourly"  # This data handler requires hourly data resolution, does it?

    def make_input_target(self):
        data, self.meta = self.load_data(self.path, self.station, self.statistics_per_var, self.sampling,
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         self.start, self.end)
        self._data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
                                      limit=self.interpolation_limit)
        self.set_inputs_and_targets()
        self.apply_filter()
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def apply_filter(self):
        raise NotImplementedError

    def create_filter_index(self) -> pd.Index:
        """Create name for filter dimension."""
        raise NotImplementedError

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables, filter.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim,
                                      self.filter_dim).copy()

    def _create_lazy_data(self):
        raise NotImplementedError

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data, self.input_data, self.target_data = list(map(f_prep, [_data, _input_data, _target_data]))


class DataHandlerFirFilterSingleStation(DataHandlerFilterSingleStation):
    """Data handler for a single station to be used by a superior data handler. Inputs are FIR filtered."""

    _requirements = remove_items(DataHandlerFilterSingleStation.requirements(), "station")
    _hash = DataHandlerFilterSingleStation._hash + ["filter_cutoff_period", "filter_order", "filter_window_type",
                                                    "_add_unfiltered"]

    DEFAULT_WINDOW_TYPE = ("kaiser", 5)
    DEFAULT_ADD_UNFILTERED = False

    def __init__(self, *args, filter_cutoff_period, filter_order, filter_window_type=DEFAULT_WINDOW_TYPE,
                 filter_add_unfiltered=DEFAULT_ADD_UNFILTERED, **kwargs):
        # self._check_sampling(**kwargs)
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.fs = self._get_fs(**kwargs)
        self.filter_cutoff_period, removed_index = self._prepare_filter_cutoff_period(filter_cutoff_period, self.fs)
        self.filter_cutoff_freq = self._period_to_freq(self.filter_cutoff_period)
        assert len(self.filter_cutoff_period) == (len(filter_order) - len(removed_index))
        self.filter_order = self._prepare_filter_order(filter_order, removed_index, self.fs)
        self.filter_window_type = filter_window_type
        self._add_unfiltered = filter_add_unfiltered
        super().__init__(*args, **kwargs)

    @staticmethod
    def _prepare_filter_order(filter_order, removed_index, fs):
        order = []
        for i, o in enumerate(filter_order):
            if i not in removed_index:
                fo = int(o * fs)
                fo = fo + 1 if fo % 2 == 0 else fo
                order.append(fo)
        return order

    @staticmethod
    def _prepare_filter_cutoff_period(filter_cutoff_period, fs):
        """Frequency must be smaller than the sampling frequency fs. Otherwise remove given cutoff period pair."""
        cutoff_tmp = (lambda x: [x] if isinstance(x, tuple) else to_list(x))(filter_cutoff_period)
        cutoff = []
        removed = []
        for i, (low, high) in enumerate(cutoff_tmp):
            low = low if (low is None or low > 2. / fs) else None
            high = high if (high is None or high > 2. / fs) else None
            if any([low, high]):
                cutoff.append((low, high))
            else:
                removed.append(i)
        return cutoff, removed

    @staticmethod
    def _period_to_freq(cutoff_p):
        return list(map(lambda x: (1. / x[0] if x[0] is not None else None, 1. / x[1] if x[1] is not None else None),
                        cutoff_p))

    @staticmethod
    def _get_fs(**kwargs):
        """Return frequency in 1/day (not Hz)"""
        sampling = kwargs.get("sampling")
        if sampling == "daily":
            return 1
        elif sampling == "hourly":
            return 24
        else:
            raise ValueError(f"Unknown sampling rate {sampling}. Only daily and hourly resolution is supported.")

    @TimeTrackingWrapper
    def apply_filter(self):
        """Apply FIR filter only on inputs."""
        fir = FIRFilter(self.input_data, self.fs, self.filter_order, self.filter_cutoff_freq, self.filter_window_type,
                        self.target_dim)
        self.fir_coeff = fir.filter_coefficients()
        fir_data = fir.filtered_data()
        if self._add_unfiltered is True:
            fir_data.append(self.input_data)
        self.input_data = xr.concat(fir_data, pd.Index(self.create_filter_index(), name=self.filter_dim))
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def create_filter_index(self) -> pd.Index:
        """
        Create name for filter dimension. Use 'high' or 'low' for high/low pass data and 'bandi' for band pass data with
        increasing numerator i (starting from 1). If 1 low, 2 band, and 1 high pass filter is used the filter index will
        become to ['low', 'band1', 'band2', 'high'].
        """
        index = []
        band_num = 1
        for (low, high) in self.filter_cutoff_period:
            if low is None:
                index.append("low")
            elif high is None:
                index.append("high")
            else:
                index.append(f"band{band_num}")
                band_num += 1
        if self._add_unfiltered:
            index.append("unfiltered")
        return pd.Index(index, name=self.filter_dim)

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data, self.fir_coeff]

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.fir_coeff = lazy_data
        super(__class__, self)._extract_lazy((_data, _meta, _input_data, _target_data))


class DataHandlerFirFilter(DefaultDataHandler):
    """Data handler using FIR filtered data."""

    data_handler = DataHandlerFirFilterSingleStation
    data_handler_transformation = DataHandlerFirFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerKzFilterSingleStation(DataHandlerFilterSingleStation):
    """Data handler for a single station to be used by a superior data handler. Inputs are kz filtered."""

    _requirements = remove_items(inspect.getfullargspec(DataHandlerFilterSingleStation).args, ["self", "station"])
    _hash = DataHandlerFilterSingleStation._hash + ["kz_filter_length", "kz_filter_iter"]

    def __init__(self, *args, kz_filter_length, kz_filter_iter, **kwargs):
        self._check_sampling(**kwargs)
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.kz_filter_length = to_list(kz_filter_length)
        self.kz_filter_iter = to_list(kz_filter_iter)
        self.cutoff_period = None
        self.cutoff_period_days = None
        super().__init__(*args, **kwargs)

    @TimeTrackingWrapper
    def apply_filter(self):
        """Apply kolmogorov zurbenko filter only on inputs."""
        kz = KZFilter(self.input_data, wl=self.kz_filter_length, itr=self.kz_filter_iter, filter_dim=self.time_dim)
        filtered_data: List[xr.DataArray] = kz.run()
        self.cutoff_period = kz.period_null()
        self.cutoff_period_days = kz.period_null_days()
        self.input_data = xr.concat(filtered_data, pd.Index(self.create_filter_index(), name=self.filter_dim))
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="74d", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def create_filter_index(self) -> pd.Index:
        """
        Round cut off periods in days and append 'res' for residuum index.

        Round small numbers (<10) to single decimal, and higher numbers to int. Transform as list of str and append
        'res' for residuum index.
        """
        index = np.round(self.cutoff_period_days, 1)
        f = lambda x: int(np.round(x)) if x >= 10 else np.round(x, 1)
        index = list(map(f, index.tolist()))
        index = list(map(lambda x: str(x) + "d", index)) + ["res"]
        return pd.Index(index, name=self.filter_dim)

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data, self.cutoff_period, self.cutoff_period_days]

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.cutoff_period, self.cutoff_period_days = lazy_data
        super(__class__, self)._extract_lazy((_data, _meta, _input_data, _target_data))


class DataHandlerKzFilter(DefaultDataHandler):
    """Data handler using kz filtered data."""

    data_handler = DataHandlerKzFilterSingleStation
    data_handler_transformation = DataHandlerKzFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerClimateFirFilterSingleStation(DataHandlerFirFilterSingleStation):
    """
    Data handler for a single station to be used by a superior data handler. Inputs are FIR filtered. In contrast to
    the simple DataHandlerFirFilterSingleStation, this data handler is centered around t0 to have no time delay. For
    values in the future (t > t0), this data handler assumes a climatological value for the low pass data and values of
    0 for all residuum components.

    :param apriori: Data to use as apriori information. This should be either a xarray dataarray containing monthly or
        any other heuristic to support the clim filter, or a list of such arrays containint heuristics for all residua
        in addition. The 2nd can be used together with apriori_type `residuum_stats` which estimates the error of the
        residuum when the clim filter should be applied with exogenous parameters. If apriori_type is None/`zeros` data
        can be provided, but this is not required in this case.
    :param apriori_type: set type of information that is provided to the clim filter. For the first low pass always a
        calculated or given statistic is used. For residuum prediction a constant value of zero is assumed if
        apriori_type is None or `zeros`, and a climatology of the residuum is used for `residuum_stats`.
    :param apriori_diurnal: use diurnal anomalies of each hour as addition to the apriori information type chosen by
        parameter apriori_type. This is only applicable for hourly resolution data.
    """

    _requirements = remove_items(DataHandlerFirFilterSingleStation.requirements(), "station")
    _hash = DataHandlerFirFilterSingleStation._hash + ["apriori_type", "apriori_sel_opts", "apriori_diurnal"]
    _store_attributes = DataHandlerFirFilterSingleStation.store_attributes() + ["apriori"]

    def __init__(self, *args, apriori=None, apriori_type=None, apriori_diurnal=False, apriori_sel_opts=None,
                 plot_path=None, **kwargs):
        self.apriori_type = apriori_type
        self.climate_filter_coeff = None  # coefficents of the used FIR filter
        self.apriori = apriori  # exogenous apriori information or None to calculate from data (endogenous)
        self.apriori_diurnal = apriori_diurnal
        self.all_apriori = None  # collection of all apriori information
        self.apriori_sel_opts = apriori_sel_opts  # ensure to separate exogenous and endogenous information
        self.plot_path = plot_path  # use this path to create insight plots
        super().__init__(*args, **kwargs)

    @TimeTrackingWrapper
    def apply_filter(self):
        """Apply FIR filter only on inputs."""
        self.apriori = self.apriori.get(str(self)) if isinstance(self.apriori, dict) else self.apriori
        climate_filter = ClimateFIRFilter(self.input_data, self.fs, self.filter_order, self.filter_cutoff_freq,
                                          self.filter_window_type, time_dim=self.time_dim, var_dim=self.target_dim,
                                          apriori_type=self.apriori_type, apriori=self.apriori,
                                          apriori_diurnal=self.apriori_diurnal, sel_opts=self.apriori_sel_opts,
                                          plot_path=self.plot_path, plot_name=str(self))
        self.climate_filter_coeff = climate_filter.filter_coefficients

        # store apriori information: store all if residuum_stat method was used, otherwise just store initial apriori
        if self.apriori_type == "residuum_stats":
            self.apriori = climate_filter.apriori_data
        else:
            self.apriori = climate_filter.initial_apriori_data
        self.all_apriori = climate_filter.apriori_data
        climate_filter_data = climate_filter.filtered_data

        # add unfiltered raw data
        if self._add_unfiltered is True:
            climate_filter_data.append(self.input_data)

        # create input data with filter index
        self.input_data = xr.concat(climate_filter_data, pd.Index(self.create_filter_index(), name=self.filter_dim))

        # this is just a code snippet to check the results of the filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def create_filter_index(self) -> pd.Index:
        """
        Round cut off periods in days and append 'res' for residuum index.

        Round small numbers (<10) to single decimal, and higher numbers to int. Transform as list of str and append
        'res' for residuum index. Add index unfiltered if the raw / unfiltered data is appended to data in addition.
        """
        index = np.round(self.filter_cutoff_period, 1)
        f = lambda x: int(np.round(x)) if x >= 10 else np.round(x, 1)
        index = list(map(f, index.tolist()))
        index = list(map(lambda x: str(x) + "d", index)) + ["res"]
        if self._add_unfiltered:
            index.append("unfiltered")
        return pd.Index(index, name=self.filter_dim)

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data, self.climate_filter_coeff,
                self.apriori, self.all_apriori]

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.climate_filter_coeff, self.apriori, self.all_apriori = lazy_data
        DataHandlerSingleStation._extract_lazy(self, (_data, _meta, _input_data, _target_data))

    @staticmethod
    def _prepare_filter_order(filter_order, removed_index, fs):
        order = []
        for i, o in enumerate(filter_order):
            if i not in removed_index:
                fo = int(o * fs)
                fo = fo + 1 if fo % 2 == 0 else fo
                order.append(fo)
        return order

    @staticmethod
    def _prepare_filter_cutoff_period(filter_cutoff_period, fs):
        """Frequency must be smaller than the sampling frequency fs. Otherwise remove given cutoff period pair."""
        cutoff = []
        removed = []
        for i, period in enumerate(to_list(filter_cutoff_period)):
            if period > 2. / fs:
                cutoff.append(period)
            else:
                removed.append(i)
        return cutoff, removed

    @staticmethod
    def _period_to_freq(cutoff_p):
        return [1. / x for x in cutoff_p]


class DataHandlerClimateFirFilter(DefaultDataHandler):
    """Data handler using climatic adjusted FIR filtered data."""

    data_handler = DataHandlerClimateFirFilterSingleStation
    data_handler_transformation = DataHandlerClimateFirFilterSingleStation
    _requirements = data_handler.requirements()
    _store_attributes = data_handler.store_attributes()
